# -*- coding: utf-8 -*-

# The order that we import these different modules is important to prevent
# circular imports errors, so we prevent isort from changing this file.
# isort: skip_file

from .fixed_composition import StructurePrediction__Python__FixedComposition
from .new_individual import StructurePrediction__Python__NewIndividual
from .variable_composition import StructurePrediction__Python__VariableComposition
from .binary_composition import StructurePrediction__Python__BinaryComposition
